#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include "valgrindifc.h"
#include "parser.h"

static struct {
  char* lastline;
  gnogrind_entry_t* lastentry;
  char* prefix;
  guint preflen;
} theParser;


static void ggparser_readline(const char* line);
static void ggparser_flush();


void
ggparser_init(int pid)
{
  char buff[64];
  theParser.lastline = NULL;
  theParser.lastentry = NULL;
  sprintf(buff,"==%d== ",pid);
  theParser.prefix = strdup(buff);
  theParser.preflen = strlen(buff);
}

void
ggparser_readbuf(const gchar* buff, guint size)
{
  const char* head;
  const char* tail;
  
  tail = head = buff;
  while(tail < (buff + size)) {
    if(*tail == '\n'){
      char* newline;
      /* found a line ... */
      if(theParser.lastline){
	newline = malloc(strlen(theParser.lastline)+(tail-head)+1);
	strcpy(newline,theParser.lastline);
	free(theParser.lastline);
	theParser.lastline = NULL;
      } else {
	newline = malloc(tail-head+1);
	newline[0]=0;
      }
      strncat(newline,head,tail-head);

      ggparser_readline(newline);
      free(newline);
      head = tail+1;
    }
    tail++;
  }
  /* some leftovers...*/
  if(head < tail) {
    char* newline;
    if(theParser.lastline){
      newline = malloc(strlen(theParser.lastline)+(tail-head)+1);
      strcpy(newline,theParser.lastline);
      free(theParser.lastline);
    } else {
      newline = malloc(tail-head+1);
      newline[0]=0;
    }
    strncat(newline,head,tail-head);
    theParser.lastline = newline;
  }
}

void
ggparser_finish()
{
  ggparser_flush();

  /* cleanup */
  free(theParser.prefix);
  theParser.prefix = NULL;
  theParser.lastentry = NULL;
  if(theParser.lastline){
    free(theParser.lastline);
  }
  theParser.lastline = NULL;
}


#define NUMOF_ENTRIES 8
#define NUMOF_DATAS   4
#define CONDJUMP_STR  "Conditional jump or move"
#define INVREAD_STR   "Invalid read"
#define INVWRITE_STR  "Invalid write"
#define MISSMATCH_STR "Mismatched "
#define SYSCALLP_STR  "Syscall param"
#define INVFREE_STR   "Invalid free()"
#define UNINIT_STR    "Use of uninitialised"
#define SUMMARY_STR   "ERROR SUMMARY:" 
#define AT_STR        "at "
#define BY_STR        "by "
#define ADDRESS_STR   "Address "
#define SUMMFREE_STR  "malloc/free:"

static void
ggparser_readline(const char* line)
{
  static struct {
    const char* prefix;
    int   preflen;
    gnogrind_entrytype_t type;
  }  entries[] = {
    {CONDJUMP_STR, 0,GG_ENTRY_CONDITION},
    {INVREAD_STR,  0,GG_ENTRY_READ},
    {INVWRITE_STR, 0,GG_ENTRY_WRITE},
    {MISSMATCH_STR,0,GG_ENTRY_MISSMATCH},
    {SYSCALLP_STR,0,GG_ENTRY_PARAM},
    {INVFREE_STR,0,GG_ENTRY_FREE},
    {UNINIT_STR,0,GG_ENTRY_UNINIT},
    {SUMMARY_STR,0,GG_ENTRY_SUMMARY}
  }, datas[] = {
    {ADDRESS_STR,0,GG_ENTRY_NONE},
    {BY_STR,     0,GG_ENTRY_NONE},
    {AT_STR,     0,GG_ENTRY_NONE},
    {SUMMFREE_STR, 0, GG_ENTRY_NONE}
  };
  int i;
  
  /*puts(line);*/
  /* initialize length once */
  if(entries[0].preflen == 0){
    for(i=0;i<NUMOF_ENTRIES;i++){
      entries[i].preflen = strlen(entries[i].prefix);
    }
    for(i=0;i<NUMOF_DATAS;i++){
      datas[i].preflen = strlen(datas[i].prefix);
    }
  }

  /* unknown line */
  if(strncmp(line,theParser.prefix,theParser.preflen)) {
    fprintf(stderr,"-E-Gnogrind: Unknown line prefix (%s)\n",line);
    return;
  }
  line += theParser.preflen;
  /* we have correct prefix ... */
  if(line[0] == '\0'){
    /* empty line is the entry separator */
    ggparser_flush();
    return;
  }
  /* check for new entry */
  for(i=0;i<NUMOF_ENTRIES;i++){
    if(!strncmp(line,entries[i].prefix,entries[i].preflen)){
      ggparser_flush();
      theParser.lastentry = gnogrind_entry_new();
      theParser.lastentry->type = entries[i].type;
      theParser.lastentry->header = strdup(line);
      return;
    }
  }
  
  /* check for data for new existing entry */
  if(theParser.lastentry) {
    if((theParser.lastentry->type != GG_ENTRY_SUMMARY && strncmp(line,"   ",3) == 0) ||
       theParser.lastentry->type == GG_ENTRY_SUMMARY) {
      if(theParser.lastentry->type != GG_ENTRY_SUMMARY) {
	line+=3;
      }
      for(i=0;i<NUMOF_DATAS;i++){
	if(!strncmp(line,datas[i].prefix,datas[i].preflen)){
	  char* d = strdup(line);
	  g_ptr_array_add(theParser.lastentry->data,d);
	  return;
	}
      }
    }
  }
  
  if(line[0]){
    fprintf(stderr,"%s\n",line);
  }
}

static void
ggparser_flush()
{
  if(theParser.lastentry){
    gnogrind_handler(GG_HANDLE_ENTRY,theParser.lastentry);
    theParser.lastentry = NULL;
  }
}

#define SUPP_BUFF 1024
static char*
ggparser_readsupp_line(FILE* file)
{
  char buff[SUPP_BUFF];
  char* line;
  while((line=fgets(buff,SUPP_BUFF,file)) != NULL){
    line = g_strstrip(line);
    if(*line == '\0' || *line == '#'){
      continue;
    }
    break;
  }
  return g_strdup(line);
}

GSList*
ggparser_readsupp(const gchar* filename)
{
  GSList* ret = NULL;
  char* line;

  FILE* f;
  if ((f=fopen(filename,"r")) == NULL) {
    return ret;
  }
  
  while((line=ggparser_readsupp_line(f)) != NULL){
    if(strcmp(line,"{") == 0){
      /* read an entry */
      char* stack;
      char* type;
      char* name;
      gnogrind_supp_t* supp;
      /* read name */
      name = ggparser_readsupp_line(f);
      if(name){
	supp = NULL;
	/* read type */
	type = ggparser_readsupp_line(f);
	if (type == NULL) break;
	{
	  int i;
	  for(i=GG_SUPP_NONE+1;i<GG_SUPP_SIZE;i++){
	    if(strcmp(type,gg_supp_types[i]) == 0){
	      supp = gnogrind_supp_new();
	      supp->name = name;
	      supp->type = i;
	      break;
	    }
	  }
	  /* unknown type specification */
	  if (i >= GG_SUPP_SIZE){
	    fprintf(stderr,"-E-Gnogrind: unknown entry \"%s\"\n",type);
	  }
	}
	/* read stack */
	if (supp == NULL) continue;

	while((stack = ggparser_readsupp_line(f)) != NULL) {
	  if(strcmp(stack,"}") == 0){
	    break;
	  }
	  {
	    int i;
	    for(i=GG_STACK_NONE+1;i<GG_STACK_SIZE;i++){
	      if(supp->type == GG_SUPP_PARAM || strncmp(gg_stack_types[i],stack,3) == 0){
		gnogrind_stack_t* s;
		s = gnogrind_stack_new();
		s->type = i;
		s->value = stack;
		g_ptr_array_add(supp->stack,s);
		break;
	      }
	    }
	    if (i >= GG_STACK_SIZE){
	      fprintf(stderr,"-E-Gnogrind: Unknown stack spec \"%s\"\n",stack);
	    }
	  }
	}
	if(stack){
	  g_free(stack);
	}
	/* add new suppression */
	if (supp){
	  ret = g_slist_append(ret,supp);
	}
      }
    } else {
      fprintf(stderr,"-E-Gnogrind: Unknown line \"%s\"\n",line);
    }
    g_free(line);
  }

  fclose(f);
  return ret;
}

gboolean
ggparser_writesupp(const gchar* filename,GSList* data)
{
  FILE* f;
  gboolean ret = TRUE;
  int i;

  if((f=fopen(filename,"w")) == NULL){
    return FALSE;
  }

  while(data){
    gnogrind_supp_t* supp = (gnogrind_supp_t*)data->data;
    fprintf(f,"{\n%s\n%s\n",supp->name,gg_supp_types[supp->type]);
    for(i = 0 ; i<supp->stack->len; i++){
      gnogrind_stack_t* stack = (gnogrind_stack_t*)g_ptr_array_index(supp->stack,i);
      fprintf(f,"\t%s\n",stack->value);
    }
    fprintf(f,"}\n");
    data = data->next;
  }

  fclose(f);
  return ret;
}
