#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gnome.h>
#include "valgrindifc.h"
#include "parser.h"

const char* gg_supp_types[GG_SUPP_SIZE] = {
  "","Value1","Value2","Value4","Value8","Addr1","Addr2","Addr4","Addr8","Cond",
  "Param","PThread","Free"
};
const char* gg_stack_types[GG_STACK_SIZE] = {
  "", "obj", "fun"
};
const char* gg_entry_types[GG_ENTRY_SIZE] = {
  "", "delete()/free() missmatch",
  "Invalid write","Invalid read",
  "Invalid free()","Conditional jump",
  "Invalid syscall param","Uninitialized value use",
  ""
};

typedef struct {
  /* valgrind members */
  gchar* appname;
  GPtrArray* appargs;
  gchar* valgrind;
  gchar* suppress;
  guint depth;
  gboolean demangle;
  gboolean limit;
  /* status members */
  gnogrind_status_t status;
  GIOChannel* pipe;
  guint       pipeid;
  int pid;
} gnogrind_settings_t;

static gnogrind_settings_t theApp;

static GIOChannel* gnogrind_execute();
static void         gnogrind_halt();
static gint gnogrind_read(GIOChannel* channel, GIOCondition cond, gpointer data);

int
gnogrind_init(int argc, char* argv[])
{
  gboolean def;
  /* read configuration file */
  theApp.valgrind = gnome_config_get_string_with_default("/gnogrind/valgrind/valgrind=/usr/bin/valgrind",&def);
  theApp.suppress = gnome_config_get_string_with_default("/gnogrind/valgrind/suppress=~/.gnogrind.supp",&def);
  theApp.appname = NULL;
  theApp.appargs = g_ptr_array_new();
  theApp.status = GG_STOPPED;
  theApp.pipe = NULL;
  theApp.pid = -1;
  theApp.depth    = gnome_config_get_int_with_default("/gnogrind/valgrind/depth=4",&def);
  theApp.demangle = gnome_config_get_bool_with_default("/gnogrind/valgrind/demangle=TRUE",&def);
  theApp.limit    = gnome_config_get_bool_with_default("/gnogrind/valgrind/limit=TRUE",&def);
  return 0;
}

void
gnogrind_finish()
{
  /* write configuration file */
  gnome_config_set_string("/gnogrind/valgrind/valgrind",theApp.valgrind);
  gnome_config_set_string("/gnogrind/valgrind/suppress",theApp.suppress);
  gnome_config_set_int("/gnogrind/valgrind/depth",theApp.depth);
  gnome_config_set_bool("/gnogrind/valgrind/demangle",theApp.demangle);
  gnome_config_set_bool("/gnogrind/valgrind/limit",theApp.limit);
  gnome_config_sync();
}

gpointer
gnogrind_status(gnogrind_status_t status)
{
  switch(status){
  case GG_RUNNING:
  case GG_STOPPED:
    return (gpointer)(theApp.status == status);
  case GG_GET_APPNAME:
    return theApp.appname;
  case GG_GET_APPARGS:
    return theApp.appargs;
  case GG_GET_VALGRIND:
    return theApp.valgrind;
  case GG_GET_DEPTH:
    return (gpointer)theApp.depth;
  case GG_GET_DEMANGLE:
    return (gpointer)theApp.demangle;
  case GG_GET_LIMIT:
    return (gpointer)theApp.limit;
  case GG_GET_SUPPRESS:
    return theApp.suppress;
  }
  return NULL;
}


int
gnogrind_control(gnogrind_control_t control,gpointer value)
{
  switch (control){
  case GG_START:
    if(theApp.status == GG_RUNNING) return 0;
    if(gnogrind_execute() == NULL){
      return -1;
    }
    theApp.status = GG_RUNNING;
    ggparser_init(theApp.pid);
    gnogrind_handler(GG_HANDLE_STATUS,(gpointer)theApp.status);
    break;
  case GG_STOP:
    if(theApp.status == GG_STOPPED) return 0;
    gnogrind_halt();
    ggparser_finish();
    theApp.status = GG_STOPPED;
    gnogrind_handler(GG_HANDLE_STATUS,(gpointer)theApp.status);
    break;
  case GG_SET_VALGRIND:
    g_free(theApp.valgrind);
    theApp.valgrind = g_strdup((gchar*)value);
    break;
  case GG_SET_APPNAME:
    if(theApp.appname){
      g_free(theApp.appname);
    }
    theApp.appname = strdup(value);
    break;
  case GG_SET_APPARGS:
    {
      int i;
      for(i = 0; i< theApp.appargs->len; i++){
	g_free(g_ptr_array_index(theApp.appargs,i));
      }
      g_ptr_array_free(theApp.appargs,TRUE);
    }
    theApp.appargs = g_ptr_array_new();
    {
      GPtrArray* newarr = (GPtrArray*)value;
      int i;
      for(i = 0; i<newarr->len; i++){
	g_ptr_array_add(theApp.appargs,
			   strdup(g_ptr_array_index(newarr,i)));
      }
    }
    break;
  case GG_SET_DEPTH:
    theApp.depth = (guint) value;
    break;
  case GG_SET_DEMANGLE:
    theApp.demangle = (gboolean) value;
    break;
  case GG_SET_LIMIT:
    theApp.limit  = (gboolean) value;
    break;
  case GG_SET_SUPPRESS:
    g_free(theApp.suppress);
    theApp.suppress = g_strdup((char*)value);
    break;
  case GG_SET_LOGFILE:
    if(theApp.status == GG_RUNNING) return 1;
    /* open a log file as if we are really running valgrind */
    {
      FILE* fid = fopen((char*)value,"r");
      int pid;
      if (fid) {
	/* read the first line and get pid from it...*/
	if(fscanf(fid,"==%d==",&pid) == 1){
	  theApp.status = GG_RUNNING;
	  theApp.pid = -1;
	  theApp.pipe = g_io_channel_unix_new(open((char*)value,O_RDONLY));
	  theApp.pipeid = g_io_add_watch(theApp.pipe,G_IO_IN|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
				 gnogrind_read,0);
	  ggparser_init(pid);
	  gnogrind_handler(GG_HANDLE_STATUS,(gpointer)theApp.status);	  
	}
	fclose(fid);
      }
    }
    break;
  default:
    /* this shouldn't happen ! */
    break;
  }
  return 0;
}

# define BUFFSIZE 256
static gint gnogrind_read(GIOChannel* channel, GIOCondition cond, gpointer data)
{
  switch(cond) {
    case G_IO_ERR:
    case G_IO_HUP:
    case G_IO_NVAL:
      /* Error happend */
      break;
    case G_IO_IN:
      /* read more data */
      {
	gchar buff[BUFFSIZE+1];
	int bread;
	GIOError err=G_IO_ERROR_AGAIN;
	
	while(err == G_IO_ERROR_AGAIN){
	  err=g_io_channel_read(channel,buff,BUFFSIZE,&bread);
	  if(bread){
	    /*fwrite(buff,bread,1,stderr);*/
	    ggparser_readbuf(buff,bread);
	  } else {
	    /* EOF ? */
	    err = G_IO_ERROR_UNKNOWN;
	  }
	}
	if (err != G_IO_ERROR_NONE){
	  break;
	}
      }
      return TRUE;
    default:
      /* this shouldn't happen */
      break;
  }

  /* some error happend or it has stopped... */
  gnogrind_control(GG_STOP,NULL);
  return FALSE;
}


GIOChannel* gnogrind_execute()
{
  int pids[2];
  GPtrArray* args = g_ptr_array_new();
  GIOChannel* ret = NULL;
  int pid;
  char pipebuf[32];
  char depthbuf[32];
  char demanglebuf[32];
  int i;
  char* delme = NULL;

  if(pipe(pids)){
    fprintf(stderr,"-E-Gnogrind: Failed to open pipes\n");
    return NULL;
  }
  sprintf(pipebuf,"--logfile-fd=%d",pids[1]);
  sprintf(depthbuf,"--num-callers=%d",theApp.depth);
  sprintf(demanglebuf,"--demangle=%s",((theApp.demangle==TRUE)? "yes" : "no"));

  /* Valgrind settings */
  g_ptr_array_add(args,theApp.valgrind);
  g_ptr_array_add(args,depthbuf);
  g_ptr_array_add(args,demanglebuf);
  /* suppressions */
  {
    struct stat dummy;
    if(!stat(theApp.suppress,&dummy)){
      delme = g_strdup_printf("--suppressions=%s",theApp.suppress);
      g_ptr_array_add(args,delme);
    }
  }
  g_ptr_array_add(args,pipebuf);

  /* application and arguments */
  g_ptr_array_add(args,theApp.appname);
  for(i=0; i < theApp.appargs->len; i++){
    g_ptr_array_add(args,g_ptr_array_index(theApp.appargs,i));
  }
  g_ptr_array_add(args,NULL);
  
  pid = fork();
  if (pid < 0) {
    fprintf(stderr,"-E-Gnogrind: Failed to fork\n");
    /* error forking */
    close(pids[0]);
    close(pids[1]);
    g_ptr_array_free(args,TRUE);
    if (delme) g_free(delme);
    return NULL;
  }
  
  if (pid == 0) {
    int i = 0;
    /* child */
    fprintf(stderr,"-I-Gnogrind: Executing");
    for(i=0;i<args->len-1;i++){
      fprintf(stderr," %s",(char*)g_ptr_array_index(args,i));
    }
    fprintf(stderr,"\n");

    if(execv(theApp.valgrind,(char**)args->pdata)){
      fprintf(stderr,"-E-Gnogrind: Failed to execute\n");
      abort();
    }
  }
  /* parent */
  g_ptr_array_free(args,TRUE);
  if(delme) g_free(delme);
  close(pids[1]);
  ret = g_io_channel_unix_new(pids[0]);
  theApp.pid = pid;
  theApp.pipe = ret;
  theApp.pipeid = g_io_add_watch(ret,G_IO_IN|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
				 gnogrind_read,0);
  return ret;
}

void gnogrind_halt()
{
  if(theApp.pipe){
    g_io_channel_close(theApp.pipe);
    g_io_channel_unref(theApp.pipe);
    theApp.pipe = NULL;
    if(theApp.pid > 0){
      kill(theApp.pid,SIGKILL);
    }
  }
}

gnogrind_entry_t*
gnogrind_entry_new()
{
  gnogrind_entry_t* ret = malloc(sizeof(gnogrind_entry_t));
  ret->type = GG_ENTRY_NONE;
  ret->data = g_ptr_array_new();
  ret->header = NULL;

  return ret;
}

void
gnogrind_entry_free(gnogrind_entry_t* entry)
{
  if(entry){
    int i = 0;
    char* dptr;
    if(entry->data->len){
      while((dptr=g_ptr_array_index(entry->data,i)) != NULL){
	i++;
	free(dptr);
      }
    }
    g_ptr_array_free(entry->data,TRUE);
    free(entry);
  }
}

gnogrind_stack_t*
gnogrind_stack_new()
{
  gnogrind_stack_t* ret = calloc(1,sizeof(gnogrind_stack_t));
  ret->type = GG_STACK_NONE;
  return ret;
}

void
gnogrind_stack_free(gnogrind_stack_t* stack)
{
  if(stack){
    g_free(stack->value);
  }
  free(stack);
}

gnogrind_supp_t*
gnogrind_supp_new()
{
  gnogrind_supp_t* ret = malloc(sizeof(gnogrind_supp_t));
  ret->type = GG_SUPP_NONE;
  ret->name = NULL;
  ret->stack = g_ptr_array_new();
  return ret;
}

void
gnogrind_supp_free(gnogrind_supp_t* supp)
{
  int i;
  for(i=0; i< supp->stack->len; i++){
    gnogrind_stack_free((gnogrind_stack_t*)g_ptr_array_index(supp->stack,i));
  }
  g_ptr_array_free(supp->stack,TRUE);
  g_free(supp->name);
  free(supp);
}
